'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = build;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var $registeredMacros = Symbol('registeredMacros');
var $macroState = Symbol('macroState');

/**
 * # Babel Macros
 */
function build(babel) {
  var t = babel.types;
  var traverse = babel.traverse;

  /**
   * A list of builtin macros.
   */

  var builtin = Object.create(null);

  var Macro = function () {
    function Macro(_ref) {
      var name = _ref.name;
      var macroBody = _ref.macroBody;
      var scope = _ref.scope;
      var state = _ref.state;

      _classCallCheck(this, Macro);

      this.name = name;
      this.macroBody = macroBody;
      this.scope = scope;
      this.state = state;
    }

    _createClass(Macro, [{
      key: 'run',
      value: function run(path, scope, state) {
        if (!this.macro) {
          this.macro = compileMacro(this.name, this.macroBody, this.scope, this.state);
        }
        return this.macro(path, scope, state);
      }
    }]);

    return Macro;
  }();

  builtin.DEFINE_MACRO = function defineMacro(path, scope, state) {
    var node = path.node;

    var id = node.arguments[0];
    if (!(id && t.isIdentifier(id))) {
      // @todo add test
      throw new Error('First argument to DEFINE_MACRO must be an identifier.');
    }
    var name = id.name;
    var macroBody = node.arguments[1];
    var subScope = path.get('arguments')[1].scope;

    scope[$registeredMacros] = scope[$registeredMacros] || {};
    scope[$registeredMacros][name] = new Macro({ name: name, macroBody: macroBody, scope: subScope, state: state });
    traverse(macroBody, visitors, subScope, state);
    path.remove();
  };

  function compileMacro(name, node, scope, state) {
    var paramNames = node.params.map(function (param) {
      return param.name;
    });
    var paramReferenceCounts = {};
    var references = Object.create(null);
    if (t.isFunction(node)) {
      traverse(node, visitors, scope, state);
      traverse(node, {
        enter: function enter(subPath) {
          var child = subPath.node;
          var parent = subPath.parent;

          if (subPath.isVariableDeclarator() || subPath.isFunctionDeclaration()) {
            references[child.id.name] = true;
          } else if (subPath.isIdentifier() && (!t.isFunction(parent) || parent.type === "ArrowFunctionExpression" && parent.body === child) && (!t.isMemberExpression(parent) || parent.object === child) && ~paramNames.indexOf(child.name)) {
            paramReferenceCounts[child.name] = paramReferenceCounts[child.name] || 0;
            paramReferenceCounts[child.name]++;
          }
        }
      }, scope);
    } else {
      // @todo add test
      throw new Error('Second argument to DEFINE_MACRO must be a FunctionExpression or ArrowFunctionExpression.');
    }
    return function (path, scope, state) {
      var cloned = _lodash2.default.cloneDeep(node);

      var _cloned$params$reduce = cloned.params.reduce(function (_ref4, id, index) {
        var _ref5 = _slicedToArray(_ref4, 2);

        var params = _ref5[0];
        var seen = _ref5[1];

        params[id.name] = {
          id: id,
          replacement: path.node.arguments[index] || t.identifier('undefined'),
          reference: null
        };
        seen[id.name] = false;
        return [params, seen];
      }, [{}, {}]);

      var _cloned$params$reduce2 = _slicedToArray(_cloned$params$reduce, 2);

      var params = _cloned$params$reduce2[0];
      var seen = _cloned$params$reduce2[1];

      var blockStack = [];
      var loopStack = [];
      var returnStatements = [];
      var hasEarlyReturn = false;

      traverse(cloned, {
        enter: function enter(subPath) {
          var child = subPath.node;
          var parent = subPath.parent;

          if (child.type === 'Identifier' && (parent.type !== "MemberExpression" || parent.object === child || parent.computed && parent.property === child)) {
            if (params[child.name]) {
              var param = params[child.name];
              if (param.replacement.type === 'Identifier' || param.replacement.type === 'Literal' || paramReferenceCounts[child.name] === 1 && param.replacement.type === 'MemberExpression') {
                subPath.replaceWith(param.replacement);
                seen[child.name] = param.replacement;
              } else {
                if (!seen[child.name]) {
                  seen[child.name] = scope.generateUidIdentifier(child.name);

                  getParentBlock(path).insertBefore([t.variableDeclaration('const', [t.variableDeclarator(seen[child.name], param.replacement)])]);
                }
                subPath.replaceWith(seen[child.name]);
              }
            } else if (references[child.name]) {
              if (!seen[child.name]) {
                seen[child.name] = scope.generateUidIdentifier(child.name);
              }
              subPath.replaceWith(seen[child.name]);
            }
          } else if (subPath.isReturnStatement()) {
            if (blockStack.length > 0) {
              hasEarlyReturn = true;
            }
            returnStatements.push([subPath, child]);
          } else if (subPath.isLoop()) {
            loopStack.push(subPath);
          } else if (subPath.isFunction()) {
            subPath.skip();
          } else if (subPath.isScope()) {
            blockStack.push(subPath);
          }
        },
        exit: function exit(path) {
          if (path.isLoop()) {
            loopStack.pop();
          } else if (path.isScope()) {
            blockStack.pop();
          }
        }
      }, scope);

      if (t.isStatement(cloned.body)) {
        (function () {
          var uid = scope.generateUidIdentifier(camelCase(name));
          var labelUid = scope.generateUidIdentifier('_' + name.toUpperCase());
          var parentBlock = getParentBlock(path);
          parentBlock.insertBefore([t.variableDeclaration('let', [t.variableDeclarator(uid)])]);
          returnStatements.forEach(function (_ref2) {
            var _ref3 = _slicedToArray(_ref2, 2);

            var path = _ref3[0];
            var child = _ref3[1];

            var isLast = child === cloned.body.body[cloned.body.body.length - 1];
            path.replaceWith(t.expressionStatement(t.assignmentExpression('=', uid, child.argument)));

            if (hasEarlyReturn && !isLast) {
              path.insertAfter(t.breakStatement(labelUid));
            }
          });

          if (hasEarlyReturn) {
            parentBlock.insertBefore(t.labeledStatement(labelUid, cloned.body));
          } else {
            parentBlock.insertBefore(cloned.body.body);
          }
          path.replaceWith(uid);
        })();
      } else {
        path.replaceWith(cloned.body);
      }
    };
  }

  function runMacro(path, macro, scope, state) {
    if (typeof macro === 'function') {
      macro(path, scope, state);
    } else {
      macro.run(path, scope, state);
    }
  }

  function getParentBlock(path) {
    while (path.parentPath.type !== 'Program' && path.parentPath && !path.parentPath.isStatementOrBlock()) {
      path = path.parentPath;
    }
    return path;
  }

  function getParentScope(path) {
    while (path.parentPath.type !== 'Program' && path.parentPath && !path.parentPath.isStatementOrBlock()) {
      path = path.parentPath;
    }
    return path.scope;
  }

  function camelCase(input) {
    return input.toLowerCase().replace(/_(.)/g, function (match, char) {
      return char.toUpperCase();
    });
  }

  function getMacro(node, scope, state) {
    if (node.type === 'CallExpression') {
      return getMacro(node.callee, scope, state);
    } else if (t.isIdentifier(node)) {
      if (state[$macroState].macrosDefined) {
        while (scope) {
          if (scope[$registeredMacros] && scope[$registeredMacros][node.name]) {
            return scope[$registeredMacros][node.name];
          }
          scope = scope.parent;
        }
      }
      if (builtin[node.name]) {
        return builtin[node.name];
      }
    } else if (t.isMemberExpression(node) && !node.computed && t.isIdentifier(node.property)) {
      return getMacro(node.property, scope, state);
    }
  }

  var visitors = {
    CallExpression: {
      enter: function enter(path, state) {
        var node = path.node;
        if (state[$macroState].macrosDefined) {
          if (node._processedByMacro) {
            return;
          }
          node._processedByMacro = true;
        }
        if (t.isMemberExpression(node.callee)) {
          // TODO temp locked for refactoring. need another idea for chaining more than 2 macros
          if (!node.callee.computed && getMacro(node.callee.object, path.scope, state) && getMacro(node.callee.property, path.scope, state)) {
            var head = node.callee.object;
            var tailId = node.callee.property;
            node.arguments.unshift(head); //TODO - so slow. meybe is ecponent slow ?
            var macro = getMacro(tailId, path.scope, state);
            if (macro) {
              runMacro(path, macro, path.scope, state);
            }
          }
        } else {
          var macro = getMacro(node.callee, path.scope, state);
          if (macro) {
            runMacro(path, macro, path.scope, state);
          }
        }
      }
    },
    Program: {
      enter: function enter(path, state) {
        state[$macroState] = {
          macrosDefined: false
        };
      },
      exit: function exit(path, state) {
        if (!state[$macroState].macrosDefined) {
          state[$macroState].macrosDefined = true;
          path.traverse(visitors, state);
        }
      }
    }
  };

  /**
   * Export the transformer.
   */
  return { visitor: visitors };
}